import torch
import lightning as L

class Dataset(L.LightningDataModule):
    def __init__(self, ds, input_vns, output_vns, train_frac=0.7, val_frac=0.1, test_frac=0.2,
                 batch_size=12, num_workers=4):
        super().__init__()
        self.ds = ds
        self.input_vns = input_vns
        self.output_vns = output_vns
        assert batch_size % 12 == 0, 'Batch size must be divisible by 12.'
        self.batch_size = batch_size
        self.num_workers = num_workers

        assert abs(train_frac + val_frac + test_frac - 1.0) < 1e-6, 'Train, val, and test fractions must sum to 1.'
        self.train_frac = train_frac
        self.val_frac = val_frac
        self.test_frac = test_frac

    def setup(self, stage=None):
        self.x = torch.stack([torch.from_numpy(self.ds[vn].values) for vn in self.input_vns], dim=1)
        self.y = torch.stack([torch.from_numpy(self.ds[vn].values) for vn in self.output_vns], dim=1)

        if 'year' in self.ds.dims:
            self.x = self.x.view(-1, 12, self.x.shape[1])  # shape: (N, 12, C_in)

        dataset = torch.utils.data.TensorDataset(self.x, self.y)

        # Compute dataset indices
        ds_idx = list(range(len(dataset)))
        test_size = int(len(dataset) * self.test_frac)
        train_valid_size = len(dataset) - test_size
        
        # Split train+val and test sets
        self.train_valid_idx = ds_idx[:train_valid_size]
        self.test_idx = ds_idx[train_valid_size:]
        
        train_valid_set = torch.utils.data.Subset(dataset, self.train_valid_idx)
        self.test_set = torch.utils.data.Subset(dataset, self.test_idx)
        
        # Further split into train and validation sets
        train_size = int(self.train_frac / (self.train_frac + self.val_frac) * train_valid_size)
        val_size = train_valid_size - train_size
        self.train_set, self.valid_set = torch.utils.data.random_split(train_valid_set, [train_size, val_size])

    def train_dataloader(self):
        return torch.utils.data.DataLoader(self.train_set, batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)

    def val_dataloader(self):
        return torch.utils.data.DataLoader(self.valid_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def test_dataloader(self):
        return torch.utils.data.DataLoader(self.test_set, batch_size=self.batch_size, num_workers=self.num_workers)



# get the version
from importlib.metadata import version
__version__ = version('deepGreen')
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

from . import utils
from . import vsl
from .model import GrowthModel
from .data import Dataset
from .trainer import Trainer
from .predictor import Predictor



import os
import lightning as L
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

from . import utils

class ParameterHistoryCallback(L.Callback):
    def __init__(self, params=['T1', 'T2', 'M1', 'M2']):
        super().__init__()
        self.params = params

    def on_train_start(self, trainer, pl_module):
        """Initialize parameter history dictionary when training starts."""
        trainer.mode = getattr(pl_module, 'mode', None)
        trainer.optimizer = getattr(pl_module, 'optimizer', None)
        trainer.lr = getattr(pl_module, 'lr', None)
        trainer.param_history = {}
        for param in self.params:
            param_value = getattr(pl_module, param, None)
            if param_value is not None:
                trainer.param_history[param] = [param_value.item()]

    def on_train_epoch_end(self, trainer, pl_module):
        """Record parameter values at the end of each training epoch."""
        for param in self.params:
            param_value = getattr(pl_module, param, None)
            if param_value is not None:
                trainer.param_history[param].append(param_value.item())

class Trainer(L.Trainer):
    def __init__(self, log_dirpath='./logs', name='IsoGen', min_epochs=1, max_epochs=1000):
        utils.p_header(f'Name: {name}')

        early_stop = L.pytorch.callbacks.early_stopping.EarlyStopping(
            monitor='valid_loss',  # Metric to monitor
            patience=10,          # Wait 3 epochs without improvement
            mode='min',          # Minimize the monitored metric (e.g., loss)
            verbose=True,        # Log when training stops
            min_delta=0.001      # Optional: Minimum change to qualify as improvement
        )

        ckpt_callback = L.pytorch.callbacks.ModelCheckpoint(
            dirpath=log_dirpath, filename=name,
        )

        # Remove the checkpoint file if it exists
        self.ckpt_fpath = os.path.join(log_dirpath, f'{name}.ckpt')
        utils.p_header(f'Checkpoint path: {os.path.abspath(self.ckpt_fpath)}')

        if os.path.exists(self.ckpt_fpath):
            os.remove(self.ckpt_fpath)
        else:
            pass

        super().__init__(
            accelerator='gpu', devices=1, strategy='auto',
            min_epochs=min_epochs, max_epochs=max_epochs,
            default_root_dir=log_dirpath, callbacks=[early_stop, ckpt_callback, ParameterHistoryCallback()]
        )

    def plot(self):
        # Determine number of subplots based on mode
        if self.mode == 'T':
            fig, ax = plt.subplots(figsize=(5, 3))
            axes = [ax]
            param_groups = [('T1', 'T2')]
        elif self.mode == 'M':
            fig, ax = plt.subplots(figsize=(5, 3))
            axes = [ax]
            param_groups = [('M1', 'M2')]
        else:
            fig, axes = plt.subplots(nrows=2, figsize=(5, 6), sharex=True)
            param_groups = [('T1', 'T2'), ('M1', 'M2')]

        for ax, (param1, param2) in zip(axes, param_groups):
            for param in [param1, param2]:
                if param in self.param_history:
                    values = self.param_history[param]
                    (line,) = ax.plot(values, marker='o', label=param)  # Capture line object
                    color = line.get_color()  # Get line color

                    # Annotate initial value
                    ax.text(-0.5, values[0], f'{values[0]:.2f}', ha='right', va='center', fontsize=10, color=color,
                            bbox=dict(facecolor='white', edgecolor=color, boxstyle='round,pad=0.3'))

                    # Annotate final value with slight vertical offset
                    x_final, y_final = len(values) - 1, values[-1]
                    ax.text(x_final*1.05, y_final, f'{param}={y_final:.2f}', ha='left', va='center', fontsize=10, color=color,
                            bbox=dict(facecolor='white', edgecolor=color, boxstyle='round,pad=0.3'))

            ax.set_ylabel('Parameter Value')
            ax.set_title(f'Parameters: {param1}, {param2}; Optimizer: {self.optimizer}, lr={self.lr}')
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))  # Integer xticks

        axes[-1].set_xlabel('Epoch')  # X-axis label only on last subplot

        return fig, axes


import torch
import colorama as ca
import numpy as np

def check_cuda():
    p_hint(f'CUDA availability: {torch.cuda.is_available()}')
    p_hint(f'Available GPUs: {torch.cuda.device_count()}')


def p_header(text):
    print(ca.Fore.CYAN + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_hint(text):
    print(ca.Fore.LIGHTBLACK_EX + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_success(text):
    print(ca.Fore.GREEN + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_fail(text):
    print(ca.Fore.RED + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_warning(text):
    print(ca.Fore.YELLOW + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)


def leakybucket(phi, da_T, da_P, Mmax=0.76, Mmin=0.01, alph=0.093, m_th=4.886, mu_th=5.8, rootd=1000, M0=0.2):
    # This is the monthly version, and the submontly is not implemented yet.
    T = da_T.values.reshape((-1, 12)).T
    P = da_P.values.reshape((-1, 12)).T
    nyrs = T.shape[1]

    # output vars
    M = np.ndarray((12, nyrs))
    potEv = np.ndarray((12, nyrs))

    # Compute normalized daylength (neglecting small difference in calculation for leap-years)
    latr = phi*np.pi/180;  # change to radians
    ndays = np.array([0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
    cdays = np.cumsum(ndays)

    # calculate mean monthly daylength (used for evapotranspiration in soil moisture calcs)
    jday = cdays[0:12] + .5*ndays[1:13]
    m_star = 1-np.tan(latr)*np.tan(23.439*np.pi/180*np.cos(jday*np.pi/182.625))
    mmm = np.empty(12)
    mmm[:] = np.nan

    for mo in range(12):
        if m_star[mo] < 0:
            mmm[mo] = 0
        elif m_star[mo] >0 and m_star[mo] < 2:
            mmm[mo] = m_star[mo]
        elif m_star[mo] > 2:
            mmm[mo] = 2


    nhrs = 24*np.arccos(1-mmm)/np.pi # the number of hours in the day in the middle of the month
    L = (ndays[1:13]/30)*(nhrs/12) # mean daylength in each month.

    for cyear in range(nyrs):    # begin cycling over years
        for t in range(12):      # begin cycling over months in a year
            # Compute potential evapotranspiration for current month after Thornthwaite:
            if T[t,cyear] < 0:
                Ep = 0
            elif T[t,cyear] >= 0 and T[t,cyear] < 26.5:
                istar = T[:,cyear]/5
                istar[istar<0] = 0

                I = np.sum(istar**1.514)
                a = (6.75e-7)*I**3 - (7.71e-5)*I**2 + (1.79e-2)*I + .49
                Ep = 16*L[t]*(10*T[t,cyear]/I)**a
            elif T[t,cyear] >= 26.5:
                Ep = -415.85 + 32.25*T[t,cyear] - .43* T[t,cyear]**2

            potEv[t,cyear] = Ep
            # Now calculate soil moisture according to the CPC Leaky Bucket model (see J. Huang et al, 1996).
            if t > 0:
                # evapotranspiration:
                Etrans = Ep*M[t-1,cyear]*rootd/(Mmax*rootd)
                # groundwater loss via percolation:
                G = mu_th*alph/(1+mu_th)*M[t-1,cyear]*rootd
                # runoff; contributions from surface flow (1st term) and subsurface (2nd term)
                R = P[t,cyear]*(M[t-1,cyear]*rootd/(Mmax*rootd))**m_th + (alph/(1+mu_th))*M[t-1,cyear]*rootd
                dWdt = P[t,cyear] - Etrans - R - G
                M[t,cyear] = M[t-1,cyear] + dWdt/rootd
            elif t == 0 and cyear > 0:
                # evapotranspiration:
                Etrans = Ep*M[11,cyear-1]*rootd/(Mmax*rootd)
                # groundwater loss via percolation:
                G = mu_th*alph/(1+mu_th)*M[11,cyear-1]*rootd
                # runoff; contributions from surface flow (1st term) and subsurface (2nd term)
                R = P[t,cyear]*(M[11,cyear-1]*rootd/(Mmax*rootd))**m_th + (alph/(1+mu_th))*M[11,cyear-1]*rootd
                dWdt = P[t,cyear] - Etrans - R - G
                M[t,cyear] = M[11,cyear-1] + dWdt/rootd
            elif t == 0 and cyear == 0:
                if M0 < 0:
                    M0 = .20
                # evapotranspiration (take initial soil moisture value to be 200 mm)
                Etrans = Ep*M0*rootd/(Mmax*rootd)
                # groundwater loss via percolation:
                G = mu_th*alph/(1+mu_th)*(M0*rootd)
                # runoff; contributions from surface flow (1st term) and subsurface (2nd term)
                R = P[t,cyear]*(M0*rootd/(Mmax*rootd))**m_th + (alph/(1+mu_th))*M0*rootd
                dWdt = P[t,cyear] - Etrans - R - G
                M[t,cyear] = M0 + dWdt/rootd

            # error-catching:
            if M[t,cyear] <= Mmin:
                M[t,cyear] = Mmin
            if M[t,cyear] >= Mmax:
                M[t,cyear] = Mmax
            if np.isnan(M[t,cyear])==1:
                M[t,cyear] = Mmin

    da_M = da_P.copy()
    da_M.name = 'sm'
    da_M.attrs['long_name'] = 'soil moisture'
    da_M.attrs['units'] = 'v/v'
    da_M.values = M.T.flatten()
    return da_M

def compute_gE(phi):
    gE = np.ones(12)
    gE[:] = np.nan

    # Compute normalized daylength (neglecting small difference in calculation for leap-years)
    latr = phi*np.pi/180;  # change to radians
    ndays = np.array([0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
    cdays = np.cumsum(ndays)
    sd = np.arcsin(np.sin(np.pi*23.5/180) * np.sin(np.pi * ((np.arange(1,366) - 80)/180)))   # solar declination
    y = -np.tan(np.ones(365)* latr) * np.tan(sd)
    y[y>=1] = 1
    y[y<=-1] = -1
    hdl = np.arccos(y)
    dtsi = (hdl* np.sin(np.ones(365)*latr)*np.sin(sd))+(np.cos(np.ones(365)*latr)*np.cos(sd)*np.sin(hdl))
    ndl=dtsi/np.max(dtsi) # normalized day length

    # calculate mean monthly daylength (used for evapotranspiration in soil moisture calcs)
    jday = cdays[0:12] + .5*ndays[1:13]
    m_star = 1-np.tan(latr)*np.tan(23.439*np.pi/180*np.cos(jday*np.pi/182.625))
    mmm = np.empty(12)
    mmm[:] = np.nan

    for mo in range(12):
        if m_star[mo] < 0:
            mmm[mo] = 0
        elif m_star[mo] >0 and m_star[mo] < 2:
            mmm[mo] = m_star[mo]
        elif m_star[mo] > 2:
            mmm[mo] = 2

    for t in range(12):
        gE[t] = np.mean(ndl[cdays[t]:cdays[t+1]])

    return gE

def nanstd(x, dim=None):
    mask = ~torch.isnan(x)
    count = mask.sum(dim=dim, keepdim=True)
    mean = torch.nanmean(x, dim=dim, keepdim=True)
    var = torch.nansum(((x - mean)**2) * mask, dim=dim, keepdim=True) / count
    return torch.sqrt(var).squeeze() if dim is not None else torch.sqrt(var)


import torch
import lightning as L
from . import utils

class GrowthModel(L.LightningModule):
    def __init__(self, T1=None, T2=None, M1=None, M2=None, mode='TMEint', phi=None, I_0=1, I_f=12,
                 optimizer='AdamW', lr=1e-2):
        super().__init__()
        # Initialize learnable parameters
        self.T1 = torch.nn.Parameter(torch.tensor(T1, requires_grad=True)) if T1 is not None else T1
        self.T2 = torch.nn.Parameter(torch.tensor(T2, requires_grad=True)) if T2 is not None else T2
        self.M1 = torch.nn.Parameter(torch.tensor(M1, requires_grad=True)) if M1 is not None else M1
        self.M2 = torch.nn.Parameter(torch.tensor(M2, requires_grad=True)) if M2 is not None else M2
        self.phi = torch.tensor(phi) if phi is not None else phi
        self.I_0 = I_0
        self.I_f = I_f
        self.criterion = torch.nn.MSELoss()
        self.lr = lr
        self.optimizer = optimizer
        self.mode = mode

    def normalize(self, V, V1, V2):
        gT = (V - V1) / (V2 - V1)
        return torch.clamp(gT, 0, 1)

    def forward(self, x):
        if self.mode == 'T':
            T = x[:, 0]
            return self.normalize(T, self.T1, self.T2)
        elif self.mode == 'M':
            M = x[:, 0]
            return self.normalize(M, self.M1, self.M2)
        elif self.mode == 'TM':
            T, M = x[:, 0], x[:, 1]
            gT = self.normalize(T, self.T1, self.T2)
            gM = self.normalize(M, self.M1, self.M2)
            return torch.min(gT, gM)
        elif self.mode == 'TME':
            T, M, gE = x[:, 0], x[:, 1], x[:, 2]
            gT = self.normalize(T, self.T1, self.T2)
            gM = self.normalize(M, self.M1, self.M2)
            res = torch.min(gT, gM) * gE
            return res
        elif self.mode == 'TMEint':
            if x.dim() == 3:
                T, M, gE = x[:, :, 0].flatten(), x[:, :, 1].flatten(), x[:, :, 2].flatten()
            elif x.dim() == 2:
                T, M, gE = x[:, 0], x[:, 1], x[:, 2]
            else:
                raise ValueError("Input tensor must be 2D or 3D.")

            gT = self.normalize(T, self.T1, self.T2)
            gM = self.normalize(M, self.M1, self.M2)
            Gr = torch.min(gT, gM) * gE
            nyrs = int(Gr.shape[0] // 12)
            Gr = Gr.view(nyrs, 12).T  # Reshape to (n_months, nyrs)
            width = torch.full((nyrs,), float('nan'))

            if self.phi > 0:  # Northern Hemisphere
                if self.I_0 < 0:
                    startmo = 12 + self.I_0
                    endmo = self.I_f
                    # First year: average previous year contribution
                    width[0] = torch.sum(Gr[0:endmo, 0]) + torch.sum(torch.mean(Gr[startmo-1:12, :], dim=1))
                    past_year = Gr[startmo-1:12, :-1]   # shape: (n_months, nyrs-1)
                    this_year = Gr[0:endmo, 1:]         # shape: (n_months, nyrs-1)
                    width[1:] = torch.sum(past_year, dim=0) + torch.sum(this_year, dim=0)
                else:
                    startmo = self.I_0
                    endmo = self.I_f
                    width = torch.sum(Gr[startmo-1:endmo, :], dim=0)

            elif self.phi < 0:  # Southern Hemisphere
                startmo = 6 + self.I_0
                endmo = self.I_f - 6
                this_year = Gr[startmo-1:12, :-1]   # shape: (n_months, nyrs-1)
                next_year = Gr[0:endmo, 1:]         # shape: (n_months, nyrs-1)
                width[:-1] = torch.sum(this_year, dim=0) + torch.sum(next_year, dim=0)
                width[-1] = torch.sum(Gr[startmo-1:12, -1]) + torch.sum(torch.mean(Gr[0:endmo, :], dim=1))

            # Standardize to get tree-ring proxy series
            res = (width - torch.nanmean(width)) / utils.nanstd(width)
            return res

    def training_step(self, batch, batch_idx):
        x, y = batch
        predictions = self(x)
        loss = self.criterion(predictions, y.flatten())
        self.log('train_loss', loss, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)

        if 'T' in self.mode:
            self.log('T1', self.T1, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
            self.log('T2', self.T2, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)

        if 'M' in self.mode or 'P' in self.mode:
            self.log('M1', self.M1, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
            self.log('M2', self.M2, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)

        # Log gradients
        for name, param in self.named_parameters():
            if param.grad is not None:
                self.log(f"{name}_grad", param.grad.abs().mean())

        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        predictions = self(x)
        loss = self.criterion(predictions, y.flatten())
        self.log('valid_loss', loss, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        predictions = self(x)
        loss = self.criterion(predictions, y.flatten())
        self.log('test_loss', loss, sync_dist=True)
        return loss
    
    def configure_optimizers(self):
        return torch.optim.__dict__[self.optimizer](self.parameters(), lr=self.lr)


from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='deepGreen',  # required
    version='2025.4.21',
    description='deepGreen: a machine learning based tree-ring width model',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Feng Zhu',
    author_email='fengzhu@ucar.edu',
    url='https://github.com/fzhu2e/deepGreen',
    packages=find_packages(),
    include_package_data=True,
    license='BSD-3',
    zip_safe=False,
    keywords=['Machine Learning', 'Tree-ring Width', 'Proxy System Model'],
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=[
        'colorama',
        'tqdm',
        'torch',
    ],
)



import os
import xarray as xr
import numpy as np
import torch
from sklearn.metrics import (
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
)

from . import utils

class Predictor:
    def __init__(self, model, data, ckpt_fpath):
        self.model = model
        self.data = data
        self.ckpt_fpath = ckpt_fpath
        self.ckpt = torch.load(self.ckpt_fpath, weights_only=True)
        self.model.load_state_dict(self.ckpt['state_dict'])

    def run(self, pred_fpath=None, truth_fpath=None):
        x = torch.stack([data[0] for data in self.data.test_set])
        y = torch.stack([data[1] for data in self.data.test_set])
        y = self.data.denormalize(y)
        y_hat = self.data.denormalize(self.model(x))

        self.ds_truth = xr.Dataset()
        self.ds_pred = xr.Dataset()

        for vn in self.data.output_vns:
            ds_template = self.data.ds[vn][self.data.test_idx[0]:self.data.test_idx[-1]+1]
            self.ds_truth[vn] = ds_template.copy()
            self.ds_pred[vn] = ds_template.copy()
            self.ds_truth[vn].values = y.squeeze(1).detach().numpy()
            self.ds_pred[vn].values = y_hat.squeeze(1).detach().numpy()

        if truth_fpath is not None:
            if os.path.exists(truth_fpath): os.remove(truth_fpath)
            dirpath = os.path.dirname(truth_fpath)
            if not os.path.exists(dirpath): os.makedirs(dirpath, exist_ok=True)
            self.ds_truth.to_netcdf(truth_fpath)
            utils.p_success(f'Truth saved at: "{truth_fpath}"')

        if pred_fpath is not None:
            if os.path.exists(pred_fpath): os.remove(pred_fpath)
            dirpath = os.path.dirname(pred_fpath)
            if not os.path.exists(dirpath): os.makedirs(dirpath, exist_ok=True)
            self.ds_pred.to_netcdf(pred_fpath)
            utils.p_success(f'Prediction saved at: "{pred_fpath}"')

    def plot_clim(self, vn):
        try:
            import x4c
        except ImportError:
            raise ImportError('x4c is required for this method. Please install x4c via: pip install x4c-exp.')

        x4c.set_style('journal', font_scale=1.0)

        ax_dict={
            'T_ann': (0, 0), 'P_ann': (0, 1),
            'T_DJF': (1, 0), 'P_DJF': (1, 1),
            'T_MAM': (2, 0), 'P_MAM': (2, 1),
            'T_JJA': (3, 0), 'P_JJA': (3, 1),
            'T_SON': (4, 0), 'P_SON': (4, 1),
        }
        proj_dict = {lb: 'Robinson' for lb in ax_dict.keys()}
        proj_kws_dict = {lb: dict(central_longitude=180) for lb in ax_dict.keys()}

        fig, ax = x4c.subplots(
            nrow=5, ncol=2,
            figsize=(10, 18),
            ax_loc=ax_dict,
            wspace=0.1, hspace=0.1,
            projs=proj_dict,
            projs_kws=proj_kws_dict,
            annotation=True, annotation_kws={'style': '', 'loc_x': 0}
        )

        ds_dict = {
            'Truth': self.ds_truth,
            'Prediction': self.ds_pred,
        }

        months_dict = {
            'DJF': [12, 1, 2],
            'MAM': [3, 4, 5],
            'JJA': [6, 7, 8],
            'SON': [9, 10, 11],
        }

        add_colorbar = False
        for name, ds in ds_dict.items():
            _, im = ds[vn].mean('time').x.plot(
                ax=ax[f'{name[0]}_ann'],
                cmap='PiYG',
                levels=np.linspace(-40, 0, 21),
                add_colorbar=add_colorbar,
                cbar_kwargs={'ticks': np.linspace(-40, 0, 11)},
                title=f'{name}: Annual Mean',
                return_im=True,
            )

            for label, months in months_dict.items():
                ds[vn].x.annualize(months=months).mean('time').x.plot(
                    ax=ax[f'{name[0]}_{label}'],
                    cmap='PiYG',
                    levels=np.linspace(-40, 0, 21),
                    add_colorbar=add_colorbar,
                    cbar_kwargs={'ticks': np.linspace(-40, 0, 11)},
                    title=f'{name}: {label}',
                )

        cbar = fig.colorbar(im, ax=list(ax.values()), orientation='horizontal', ticks=np.linspace(-40, 0, 11), aspect=15, pad=0.02, shrink=0.5)
        cbar.set_label(r'$\delta^{18}$O$_p$ [‰]')
        return fig, ax


    def valid_metric(self, vn, metric='R2', **plot_kws):
        try:
            import x4c
        except ImportError:
            raise ImportError('x4c is required for this method. Please install x4c via: pip install x4c-exp.')

        metric_dict = {
            'R2': r2_score,
            'MAE': mean_squared_error,
            'MSE': mean_squared_error,
            'RMSE': mean_squared_error,
            'MAE': mean_absolute_error,
            'MAPE': mean_absolute_percentage_error,
        }

        x4c.set_style('journal', font_scale=1.0)
        valid_map = xr.apply_ufunc(
            metric_dict[metric],
            self.ds_truth[vn],
            self.ds_pred[vn],
            input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
            vectorize=True,
        )
        if metric == 'RMSE': valid_map = np.sqrt(valid_map)

        _plot_kws = {
            'R2': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': '$R^2$'},
                title='$R^2$(Prediction, Truth)',
                extend='min',
            ),
            'MSE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': 'MSE'},
                title='MSE(Prediction, Truth)',
                extend='both',
            ),
            'RMSE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': 'RMSE'},
                title='RMSE(Prediction, Truth)',
                extend='both',
            ),
            'MAE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': 'MAE'},
                title='MAE(Prediction, Truth)',
                extend='both',
            ),
            'MAPE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': 'MAPE'},
                title='MAPE(Prediction, Truth)',
                extend='both',
            ),
        }
        _plot_kws[metric].update(plot_kws)
        fig, ax = valid_map.x.plot(
            **_plot_kws[metric],
        )

        return fig, ax